#!/usr/bin/env python3.9

# (C) 2008 Norbert Nemec
# This file is part of the CASINO distribution.
# Permission is given to use the script along with the CASINO program and modify
# it for personal use.

import inspect
import sys

import numpy as np

F2P_bool = {'.false.': False, '.true.': True}
P2F_bool = {False: '.false.', True: '.true.'}

num_orbs_per_shelltype = np.array([0, 1, 4, 3, 5, 7, 9])


def mapunion(a, b):
    res = a.copy()
    res.update(b)
    return res


def lineno():
    """Returns the current line number in our program."""
    return inspect.currentframe().f_back.f_lineno


def weave_inline(support_code, code, dict, defs=[]):
    try:
        import weave
    except ImportError:
        print('This program requires the weave library, which could not be')
        print('found.')
        sys.exit()
    weave.inline(
        headers=['<cstdlib>', '<cmath>'],
        support_code='\n'.join(['#define ' + d for d in defs] + [support_code]),
        code='\n'.join(['// #define ' + d for d in defs] + [code]),
        arg_names=dict.keys(),
        local_dict=dict,
        type_converters=weave.converters.blitz,
        compiler='gcc',
        extra_compile_args=['-Wno-all'],
        verbose=1,
    )


support_code = (
    r"""
#line """
    + '%i' % (lineno() + 1)
    + r""" "stowfn.py"
const double sto_exp_cutoff = 746.0;
const int num_poly_in_shell_type[] = { 0, 1, 4, 3, 5, 7, 9 };
const int first_poly_in_shell_type[] = { 0, 0, 0, 1, 4, 9, 16 };
const double pi = 3.14159265358979323846;

const int polypow[25] = {
    0,
    1,1,1,
    2,2,2,2,2,
    3,3,3,3,3,3,3,
    4,4,4,4,4,4,4,4,4
};

double factorial(int N) {
    if(N<=1) return 1.0;
    else     return N*factorial(N-1);
}

template <typename T>
inline blitz::Array<T,1> vec(const T a1,const T a2,const T a3)
{blitz::Array<T,1> res(3); res=a1,a2,a3; return res; }
"""
)

eval_code = (
    r"""
#line """
    + '%i' % (lineno() + 1)
    + r""" "stowfn.py"
blitz::Range all = blitz::Range::all();

blitz::Array<double,1> poly(25), phi(9);
#ifdef CALC_DERIVS
blitz::Array<double,2> dpoly(3,25), dphi(3,9);
blitz::Array<double,1> ddphi(9);
#endif

#ifdef EVAL_MOLORBS
blitz::Range M(0,num_molorbs);

val = 0.0;
#ifdef CALC_DERIVS
grad = 0.0;
lap = 0.0;
#endif
#endif

blitz::Array<double,1> r(blitz::Range(-1,max_order_r));

for(int pt=0;pt<num_points;pt++) {
  int n_shell=0;
  int n_atorb=0;

  for(int centre=0;centre<num_centres;centre++) {
    double x=pos(0,pt)-centrepos(centre,0);
    double y=pos(1,pt)-centrepos(centre,1);
    double z=pos(2,pt)-centrepos(centre,2);
    double xx=x*x;
    double yy=y*y;
    double zz=z*z;

    r(2)=xx+yy+zz;
    r(1)=sqrt(r(2));
    for(int i=3;i<=max_order_r_on_centre(centre);i++)
      r(i)=r(i-1)*r(1);

    if(max_order_r_on_centre(centre)>0) {
      r(0)=1.0;
      r(-1)=1.0/r(1);
    }

    double xnorm=x*r(-1);
    double ynorm=y*r(-1);
    double znorm=z*r(-1);

    poly(0) = 1;
    poly(1) = x;
    poly(2) = y;
    poly(3) = z;

#ifdef CALC_DERIVS
    dpoly(all,0) = vec(0.,0.,0.);
    dpoly(all,1) = vec(1.,0.,0.);
    dpoly(all,2) = vec(0.,1.,0.);
    dpoly(all,3) = vec(0.,0.,1.);
#endif

    if(max_shell_type_on_centre(centre)>=4) {
      // there are d and/or higher shells
      double xy = x*y;
      double yz = y*z;
      double zx = z*x;

      // polynomials for d shells
      poly(4)=xy;
      poly(5)=yz;
      poly(6)=zx;
      poly(7)=3*zz-r(2);
      poly(8)=xx-yy;

#ifdef CALC_DERIVS
      dpoly(all,4) = vec(   y,   x,  0.);
      dpoly(all,5) = vec(   0.,   z,  y);
      dpoly(all,6) = vec(   z,   0.,  x);
      dpoly(all,7) = vec(-2*x,-2*y,4*z);
      dpoly(all,8) = vec( 2*x,-2*y,  0.);
#endif

      if(max_shell_type_on_centre(centre) >= 5) {
        // there are f and/or higher shells
        double t1 = 5*zz-r(2);

        poly( 9)=(2*zz-3*(xx+yy))*z;  // (2*zz-3*(xx+yy))*z
        poly(10)=t1*x;                // (4*zz-(xx+yy))*x
        poly(11)=t1*y;                // (4*zz-(xx+yy))*y
        poly(12)=(xx-yy)*z;
        poly(13)=xy*z;
        poly(14)=(xx-3.0*yy)*x;
        poly(15)=(3.0*xx-yy)*y;

#ifdef CALC_DERIVS
        dpoly(all, 9) = vec(-6*zx,-6*yz,6*zz-3*(xx+yy));
        dpoly(all,10) = vec( 4*zz-yy-3*xx,-2*xy,8*zx);
        dpoly(all,11) = vec(-2*xy,4*zz-xx-3*yy,8*yz);
        dpoly(all,12) = vec( 2*zx,-2*yz,xx-yy);
        dpoly(all,13) = vec( yz,zx,xy);
        dpoly(all,14) = vec( 3*xx-3*yy,-6*xy,0.);
        dpoly(all,15) = vec( 6*xy,3*xx-3*yy,0.);
#endif

        if(max_shell_type_on_centre(centre) >= 6) {
          // there are g shells

          double xx_yy3=xx-3*yy;
          double xx3_yy=3*xx-yy;
          double xx_yy=xx-yy;
          double zz5=5*zz;
          double zz7=7*zz;
          double rr3=3*r(2);
          double zz7_rr=zz7-r(2);
          double zz7_rr3=zz7-rr3;

          poly(16)=zz5*(zz7_rr3) - (zz5 - r(2))*rr3; // 35zzzz-30zzrr+3rrrr
          poly(17)=zx*(zz7_rr3);                     // xz(7zz-3rr)
          poly(18)=yz*(zz7_rr3);                     // yz(7zz-3rr)
          poly(19)=(xx_yy)*(zz7_rr);                 // (xx-yy)(7zz-rr)
          poly(20)=xy*(zz7_rr);                      // xy(7zz-rr)
          poly(21)=zx*(xx_yy3);                      // xz(xx-3yy)
          poly(22)=yz*(xx3_yy);                      // yz(3xx-yy)
          poly(23)=xx*(xx_yy3) - yy*(xx3_yy);        // xxxx-6xxyy+yyyy
          poly(24)=xy*(xx_yy);                       // xxxy-xyyy


#ifdef CALC_DERIVS
//          dpoly(all, 16) = -60zzx+12rrx,   -60zzy+12rry,   140zzz-60zrr-60zzz+12rrz
//          dpoly(all, 17) = 7zzz-3zrr-6zxx, -6xyz,          21xzz-3xrr-6xzz
//          dpoly(all, 18) = -6xyz,          7zzz-3zrr-6zyy, 21yzz-3yrr-6yzz
//          dpoly(all, 19) = 14xzz-2xrr-2xxx+2xyy, ...
//          dpoly(all, 20) =
//          dpoly(all, 21) =
//          dpoly(all, 22) =
//          dpoly(all, 23) =
//          dpoly(all, 24) =

          throw "gradient and laplacian of g orbitals not yet implemented";
#endif
        }
      }
    }

//std::cout << "poly = "<< poly << "\n";
//#ifdef CALC_DERIVS
//std::cout << "dpoly = "<< dpoly << "\n";
//#endif

    for(int shell=0; shell<num_shells_on_centre(centre);shell++,n_shell++) {
      double zeta_rabs=zeta(n_shell)*r(1);
      if(zeta_rabs>sto_exp_cutoff) {
        n_atorb += num_poly_in_shell_type[shelltype(n_shell)];
        continue; // to next shell
      }
      double exp_zeta_rabs=exp(-zeta_rabs);
      blitz::Range A(first_poly_in_shell_type[shelltype(n_shell)],first_poly_in_shell_type[shelltype(n_shell)]+num_poly_in_shell_type[shelltype(n_shell)]-1);
      blitz::Range X(0,num_poly_in_shell_type[shelltype(n_shell)]-1);
      phi(X) = poly(A)*exp_zeta_rabs;
#ifdef CALC_DERIVS
      dphi(0,X) = dpoly(0,A)*exp_zeta_rabs-zeta(n_shell)*xnorm*phi(X);
      dphi(1,X) = dpoly(1,A)*exp_zeta_rabs-zeta(n_shell)*ynorm*phi(X);
      dphi(2,X) = dpoly(2,A)*exp_zeta_rabs-zeta(n_shell)*znorm*phi(X);
      ddphi(X) = (
          -2*zeta(n_shell)*r(-1)*(x*dpoly(0,A)+y*dpoly(1,A)+z*dpoly(2,A)+poly(A))
          +zeta(n_shell)*zeta(n_shell)*poly(A)
        )*exp_zeta_rabs;
#endif

      int N=order_r_in_shell(n_shell);

      if(N==0) {

#ifdef EVAL_ATORBS
        for(int pl=0;pl<num_poly_in_shell_type[shelltype(n_shell)];pl++) {
          atorbs(pt,n_atorb)=phi(pl);
          n_atorb++;
        }
#endif // EVAL_ATORBS

#ifdef EVAL_MOLORBS
        for(int pl=0;pl<num_poly_in_shell_type[shelltype(n_shell)];pl++) {
          val(pt,all)+=coeff_norm(all,n_atorb)*phi(pl);
#ifdef CALC_DERIVS
          grad(0,pt,all)+=coeff_norm(all,n_atorb)*dphi(0,pl);
          grad(1,pt,all)+=coeff_norm(all,n_atorb)*dphi(1,pl);
          grad(2,pt,all)+=coeff_norm(all,n_atorb)*dphi(2,pl);
          lap(pt,all)+=coeff_norm(all,n_atorb)*ddphi(pl);
#endif
          n_atorb++;
        }
#endif // EVAL_MOLORBS

      } else {

#ifdef EVAL_ATORBS
        for(int pl=0;pl<num_poly_in_shell_type[shelltype(n_shell)];pl++) {
          atorbs(pt,n_atorb)=r(N)*phi(pl);
          n_atorb++;
        }
#endif // EVAL_ATORBS

#ifdef EVAL_MOLORBS
        for(int pl=0;pl<num_poly_in_shell_type[shelltype(n_shell)];pl++) {
          val(pt,all)+=coeff_norm(all,n_atorb)*r(N)*phi(pl);
#ifdef CALC_DERIVS
          grad(0,pt,all)+=coeff_norm(all,n_atorb)*(N*x*r(N-2)*phi(pl)+r(N)*dphi(0,pl));
          grad(1,pt,all)+=coeff_norm(all,n_atorb)*(N*y*r(N-2)*phi(pl)+r(N)*dphi(1,pl));
          grad(2,pt,all)+=coeff_norm(all,n_atorb)*(N*z*r(N-2)*phi(pl)+r(N)*dphi(2,pl));
          lap(pt,all)+=coeff_norm(all,n_atorb)*
            (N*(N+1)*r(N-2)*phi(pl)+2*N*r(N-2)*(x*dphi(0,pl)+y*dphi(1,pl)+z*dphi(2,pl))+r(N)*ddphi(pl));
#endif
          n_atorb++;
        }
#endif // EVAL_MOLORBS

      }

    } // shell
  } // centre
} // pos
"""
)


norm_code = (
    r"""
#line """
    + '%i' % (lineno() + 1)
    + r""" "stowfn.py"
int n_shell=0;
int n_atorb=0;
double polynorm[25];

polynorm[0] = sqrt(1./(4.*pi)); // 1
polynorm[1] = sqrt(3./(4.*pi)); // x
polynorm[2] = sqrt(3./(4.*pi)); // y
polynorm[3] = sqrt(3./(4.*pi)); // z

polynorm[4] = .5*sqrt(15./pi); // xy
polynorm[5] = .5*sqrt(15./pi); // yz
polynorm[6] = .5*sqrt(15./pi); // zx
polynorm[7] = .25*sqrt(5./pi); // 3*zz-r(2);
polynorm[8] = .25*sqrt(15./pi); // xx-yy;

polynorm[ 9] = .25*sqrt(7./pi); // (2*zz-3*(xx+yy))*z;
polynorm[10] = .25*sqrt(10.5/pi); // (4*zz-(xx+yy))*x;
polynorm[11] = .25*sqrt(10.5/pi); // (4*zz-(xx+yy))*y;
polynorm[12] = .25*sqrt(105./pi); // (xx-yy)*z;
polynorm[13] = .5*sqrt(105./pi); // xy*z;
polynorm[14] = .25*sqrt(17.5/pi); // (xx-3.0*yy)*x;
polynorm[15] = .25*sqrt(17.5/pi); // (3.0*xx-yy)*y;

polynorm[16] = .1875*sqrt(1./pi); // 35zzzz-30zzrr+3rrrr
polynorm[17] = .75*sqrt(2.5/pi); // xz(7zz-3rr)
polynorm[18] = .75*sqrt(2.5/pi); // yz(7zz-3rr)
polynorm[19] = .375*sqrt(5./pi); // (xx-yy)(7zz-rr)
polynorm[20] = .75*sqrt(5./pi); // xy(7zz-rr)
polynorm[21] = .75*sqrt(17.5/pi); // xz(xx-3yy)
polynorm[22] = .75*sqrt(17.5/pi); // yz(3xx-yy)
polynorm[23] = .1875*sqrt(35./pi); // xxxx-6xxyy+yyyy
polynorm[24] = .75*sqrt(35./pi); // xxxy-xyyy

for(int centre=0; centre<num_centres;centre++) {
    for(int shell=0; shell<num_shells_on_centre(centre); shell++,n_shell++) {
        for(int pl=first_poly_in_shell_type[shelltype(n_shell)];
            pl < first_poly_in_shell_type[shelltype(n_shell)]+num_poly_in_shell_type[shelltype(n_shell)];
            pl++, n_atorb++) {
            int n = polypow[pl] + order_r_in_shell(n_shell) + 1;
            norm(n_atorb)=polynorm[pl] * pow(2*zeta(n_shell),n) * sqrt(2*zeta(n_shell)/factorial(2*n));
        } // pl
    } // shell
} // centre
"""
)


class stowfn:
    def __init__(self, fname=None):
        if fname is not None:
            self.readfile(fname)
        else:
            self.initempty()

    def initempty(self):
        pass

    def readfile(self, fname):
        f = open(fname, 'r')

        def readline():
            return f.readline()

        def readstr():
            return readline().strip()

        def readint():
            return int(readstr())

        def readfloat():
            return float(readstr())

        def readfloats(N):
            res = []
            while len(res) < N:
                l = readline()
                res += [float(l[i : i + 20]) for i in range(0, len(l) - 1, 20)]
            assert len(res) == N
            return np.array(res)

        def readints(N):
            res = []
            while len(res) < N:
                res += [int(n) for n in readline().split()]
            assert len(res) == N
            return np.array(res)

        def readbool():
            return F2P_bool[readstr()]

        def skipline(text=''):
            l = f.readline()
            assert l == text + '\n'

        self.title = readstr()
        skipline()
        skipline('BASIC INFO')
        skipline('----------')
        skipline('Generated by:')
        self.code = readstr()
        skipline('Periodicity:')
        self.periodicity = readint()
        skipline('Spin unrestricted:')
        self.spin_unrestricted = readbool()
        skipline('Nuclear repulsion energy (au/atom):')
        self.nuclear_repulsion_energy = readfloat()
        skipline('Number of electrons')
        self.num_elec = readint()
        skipline()
        skipline('GEOMETRY')
        skipline('--------')
        skipline('Number of atoms')
        self.num_atom = readint()
        skipline('Atomic positions (au)')
        self.atompos = readfloats(self.num_atom * 3).reshape((self.num_atom, 3))
        skipline('Atomic numbers for each atom')
        self.atomnum = readints(self.num_atom)
        skipline('Valence charges for each atom')
        self.atomcharge = readfloats(self.num_atom)
        skipline()
        skipline('BASIS SET')
        skipline('---------')
        skipline('Number of STO centres')
        self.num_centres = readint()
        skipline('Position of each centre (au)')
        self.centrepos = readfloats(self.num_centres * 3).reshape((self.num_centres, 3))
        skipline('Number of shells')
        self.num_shells = readint()
        skipline('Sequence number of first shell on each centre')
        self.idx_first_shell_on_centre = np.array(list(readints(self.num_centres) - 1) + [self.num_shells])
        skipline('Code for shell types (s/sp/p/d/f/g 1/2/3/4/5/6)')
        self.shelltype = readints(self.num_shells)
        skipline('Order of radial prefactor r in each shell')
        self.order_r_in_shell = readints(self.num_shells)
        skipline('Exponent in each STO shell')
        self.zeta = readfloats(self.num_shells)
        skipline("Number of basis functions ('AO')")
        self.num_atorbs = readint()
        skipline("Number of molecular orbitals ('MO')")
        self.num_molorbs = readints(1 + self.spin_unrestricted)
        skipline()

        assert self.idx_first_shell_on_centre[-1] == self.num_shells
        assert self.idx_first_shell_on_centre[0] == 0

        self.num_shells_on_centre = self.idx_first_shell_on_centre[1:] - self.idx_first_shell_on_centre[:-1]
        self.max_order_r_on_centre = np.array(
            [
                max(2, self.order_r_in_shell[self.idx_first_shell_on_centre[i] : self.idx_first_shell_on_centre[i + 1]].max())
                for i in range(self.num_centres)
            ]
        )
        self.max_order_r = self.max_order_r_on_centre.max()
        self.max_shell_type_on_centre = np.array(
            [self.shelltype[self.idx_first_shell_on_centre[i] : self.idx_first_shell_on_centre[i + 1]].max() for i in range(self.num_centres)]
        )
        assert all(self.num_shells_on_centre > 0)
        assert sum(num_orbs_per_shelltype[self.shelltype]) == self.num_atorbs

        skipline('MULTIDETERMINANT INFORMATION')
        skipline('----------------------------')
        skipline('GS')
        skipline()

        line = readline()
        if line == 'ORBITAL COEFFICIENTS (normalized AO)\n':
            skipline('------------------------------------')
            self.coeff = [readfloats(self.num_molorbs[0] * self.num_atorbs).reshape((self.num_molorbs[0], self.num_atorbs))]
            if self.spin_unrestricted:
                self.coeff += [readfloats(self.num_molorbs[1] * self.num_atorbs).reshape((self.num_molorbs[1], self.num_atorbs))]
            self.coeff_norm = [c[:, :] * self.get_norm()[None, :] for c in self.coeff]
            skipline()
            self.footer = f.readlines()
        else:
            self.coeff = [np.zeros((self.num_molorbs[0], self.num_atorbs))]
            if self.spin_unrestricted:
                self.coeff += [np.zeros((self.num_molorbs[1], self.num_atorbs))]
            self.coeff_norm = [c[:, :] * 0.0 for c in self.coeff]
            self.footer = [line] + f.readlines()
        f.close()

    def check_and_normalize(self):
        self.title = str(self.title)
        self.code = str(self.code)
        self.periodicity = int(self.periodicity)
        self.spin_unrestricted = bool(self.spin_unrestricted)
        self.nuclear_repulsion_energy = float(self.nuclear_repulsion_energy)
        self.num_elec = int(self.num_elec)
        self.num_atom = int(self.num_atom)
        assert self.atompos.shape == (self.num_atom, 3)
        assert np.issubdtype(self.atompos.dtype, float)
        assert self.atomnum.shape == (self.num_atom,)
        assert np.issubdtype(self.atomnum.dtype, int)
        assert self.atomcharge.shape == (self.num_atom,)
        assert np.issubdtype(self.atomcharge.dtype, float)
        self.num_centres = int(self.num_centres)
        assert self.centrepos.shape == (self.num_centres, 3)
        assert np.issubdtype(self.centrepos.dtype, float)
        self.num_shells = int(self.num_shells)
        assert self.idx_first_shell_on_centre.shape == (self.num_centres + 1,)
        assert np.issubdtype(self.idx_first_shell_on_centre.dtype, int)
        assert self.shelltype.shape == (self.num_shells,)
        assert np.issubdtype(self.shelltype.dtype, int)
        assert self.order_r_in_shell.shape == (self.num_shells,)
        assert np.issubdtype(self.order_r_in_shell.dtype, int)
        assert self.zeta.shape == (self.num_shells,)
        assert np.issubdtype(self.zeta.dtype, float)
        self.num_atorbs = int(self.num_atorbs)
        assert self.num_molorbs.shape == (1 + self.spin_unrestricted,)

        assert self.idx_first_shell_on_centre[-1] == self.num_shells
        assert self.idx_first_shell_on_centre[0] == 0

        self.num_shells_on_centre = self.idx_first_shell_on_centre[1:] - self.idx_first_shell_on_centre[:-1]
        self.max_order_r_on_centre = np.array(
            [
                max(2, self.order_r_in_shell[self.idx_first_shell_on_centre[i] : self.idx_first_shell_on_centre[i + 1]].max())
                for i in range(self.num_centres)
            ]
        )
        self.max_order_r = self.max_order_r_on_centre.max()
        self.max_shell_type_on_centre = np.array(
            [self.shelltype[self.idx_first_shell_on_centre[i] : self.idx_first_shell_on_centre[i + 1]].max() for i in range(self.num_centres)]
        )
        assert all(self.num_shells_on_centre > 0)
        assert sum(num_orbs_per_shelltype[self.shelltype]) == self.num_atorbs

        self.num_spins = 1 + self.spin_unrestricted
        assert len(self.coeff) == self.num_spins
        for sp in range(self.num_spins):
            assert self.coeff[sp].shape == (self.num_molorbs[sp], self.num_atorbs)
            assert self.coeff[sp].dtype == np.float64
        self.coeff_norm = [self.coeff[sp][:, :] * self.get_norm()[None, :] for sp in range(self.num_spins)]

    def writefile(self, fname):
        f = open(fname, 'w')

        def writeline(l=''):
            f.write(l + '\n')

        def writestr(s):
            writeline(' ' + s)

        def writeint(i):
            writestr('%i' % i)

        def writefloat(f):
            writestr('%.15f' % f)

        def writefloats(F, num_cols=4):
            N = len(F)
            for c in range(0, N - num_cols + 1, num_cols):
                writeline(('% .13E' * num_cols) % tuple(F[c : c + num_cols]))
            if N % num_cols:
                writeline(('% .13E' * (N % num_cols)) % tuple(F[-(N % num_cols) :]))

        def writeints(I, num_cols=8):
            N = len(I)
            for c in range(0, N - num_cols + 1, num_cols):
                writeline(('% 10i' * num_cols) % tuple(I[c : c + num_cols]))
            if N % num_cols:
                writeline(('% 10i' * (N % num_cols)) % tuple(I[-(N % num_cols) :]))

        def writebool(b):
            return writestr(P2F_bool[b])

        writeline(self.title)
        writeline()
        writeline('BASIC INFO')
        writeline('----------')
        writeline('Generated by:')
        writestr(self.code)
        writeline('Periodicity:')
        writeint(self.periodicity)
        writeline('Spin unrestricted:')
        writebool(self.spin_unrestricted)
        writeline('Nuclear repulsion energy (au/atom):')
        writefloat(self.nuclear_repulsion_energy)
        writeline('Number of electrons')
        writeint(self.num_elec)
        writeline()
        writeline('GEOMETRY')
        writeline('--------')
        writeline('Number of atoms')
        writeint(self.num_atom)
        writeline('Atomic positions (au)')
        writefloats(self.atompos.reshape((self.num_atom * 3)), num_cols=3)
        writeline('Atomic numbers for each atom')
        writeints(self.atomnum)
        writeline('Valence charges for each atom')
        writefloats(self.atomcharge)
        writeline()
        writeline('BASIS SET')
        writeline('---------')
        writeline('Number of STO centres')
        writeint(self.num_centres)
        writeline('Position of each centre (au)')
        writefloats(self.centrepos.reshape((self.num_centres * 3)), num_cols=3)
        writeline('Number of shells')
        writeint(self.num_shells)
        writeline('Sequence number of first shell on each centre')
        writeints(self.idx_first_shell_on_centre[:-1] + 1)
        writeline('Code for shell types (s/sp/p/d/f/g 1/2/3/4/5/6)')
        writeints(self.shelltype)
        writeline('Order of radial prefactor r in each shell')
        writeints(self.order_r_in_shell)
        writeline('Exponent in each STO shell')
        writefloats(self.zeta)
        writeline("Number of basis functions ('AO')")
        writeint(self.num_atorbs)
        writeline("Number of molecular orbitals ('MO')")
        writeints(self.num_molorbs)
        writeline()
        writeline('MULTIDETERMINANT INFORMATION')
        writeline('----------------------------')
        writeline('GS')
        writeline()

        if hasattr(self, 'coeff_norm'):
            writeline('ORBITAL COEFFICIENTS (normalized AO)')
            writeline('------------------------------------')
            coeff = self.coeff_norm[0][:, :] / self.get_norm()[None, :]
            writefloats(coeff.reshape((self.num_molorbs[0] * self.num_atorbs)))
            if self.spin_unrestricted:
                coeff = self.coeff_norm[1][:, :] / self.get_norm()[None, :]
                writefloats(coeff.reshape((self.num_molorbs[1] * self.num_atorbs)))
            writeline()
        elif hasattr(self, 'coeff'):
            writeline('ORBITAL COEFFICIENTS (normalized AO)')
            writeline('------------------------------------')
            writefloats(self.coeff[0].reshape((self.num_molorbs[0] * self.num_atorbs)))
            if self.spin_unrestricted:
                writefloats(self.coeff[1].reshape((self.num_molorbs[1] * self.num_atorbs)))
            writeline()

        f.writelines(self.footer)
        f.close()

    def read_molorbmods(self, fname='correlation.data'):
        l = [l.strip() for l in open(fname, 'r').readlines()]
        start = l.index('START MOLORBMODS') + 3
        end = l.index('END MOLORBMODS')
        while start < end:
            if l[start] == 'START MOLECULAR ORBITAL COEFFICIENTS':
                endcoeff = l.index('END MOLECULAR ORBITAL COEFFICIENTS')
                # moc = [float(c.split()[0]) for c in l[start + 2 : endcoeff]]
                start = endcoeff + 1
            elif l[start] == 'START STO EXPONENT ZETAS':
                endzeta = l.index('END STO EXPONENT ZETAS')
                # zet = [float(c.split()[0]) for c in l[start + 2 : endzeta]]
                start = endzeta + 1
            else:
                raise "unknown block starting with '" + l[start] + "'"

    def eval_molorbs(self, pos, spin=0):
        """Evaluate molecular orbitals at given positions.

        Args:
            pos (numpy.ndarray): Array of shape (3, num_points) with Cartesian coordinates.
            spin (int, optional): Spin index (0 for alpha, 1 for beta if unrestricted).

        Returns:
            numpy.ndarray: Values of molecular orbitals (num_points, num_molorbs).
        """
        num_points = pos.shape[1]
        assert pos.shape == (3, num_points)
        num_molorbs = self.num_molorbs[spin]
        val = np.zeros((num_points, num_molorbs))
        coeff_norm = self.coeff_norm[spin]
        dict = mapunion(self.__dict__, locals())
        weave_inline(support_code, eval_code, dict, ['EVAL_MOLORBS'])
        return val

    def eval_molorb_derivs(self, pos, spin=0):
        """Evaluate molecular orbitals, gradients, and Laplacians.

        Args:
            pos (numpy.ndarray): Array of shape (3, num_points) with Cartesian coordinates.
            spin (int, optional): Spin index (0 for alpha, 1 for beta if unrestricted).

        Returns:
            tuple:
                - val (numpy.ndarray): Orbital values (num_points, num_molorbs).
                - grad (numpy.ndarray): Orbital gradients (3, num_points, num_molorbs).
                - lap (numpy.ndarray): Orbital Laplacians (num_points, num_molorbs).
        """
        num_points = pos.shape[1]
        assert pos.shape == (3, num_points)
        num_molorbs = self.num_molorbs[spin]
        val = np.zeros((num_points, num_molorbs))
        grad = np.zeros((3, num_points, num_molorbs))
        lap = np.zeros((num_points, num_molorbs))
        coeff_norm = self.coeff_norm[spin]
        dict = mapunion(self.__dict__, locals())
        weave_inline(support_code, eval_code, dict, ['EVAL_MOLORBS', 'CALC_DERIVS'])
        return val, grad, lap

    def eval_atorbs(self, pos):
        """Evaluate atomic orbitals (AOs) at given positions.

        Args:
            pos (numpy.ndarray): Array of shape (3, num_points) with Cartesian coordinates.

        Returns:
            numpy.ndarray: AO values (num_points, num_atorbs).
        """
        num_points = pos.shape[1]
        assert pos.shape == (3, num_points)
        atorbs = np.zeros((num_points, self.num_atorbs))
        dict = mapunion(self.__dict__, locals())
        weave_inline(support_code, eval_code, dict, ['EVAL_ATORBS'])
        return atorbs

    def get_norm(self):
        """Compute normalization factors for atomic orbitals.

        Returns:
            numpy.ndarray: Norms of each AO (num_atorbs,).
        """
        norm = np.zeros((self.num_atorbs,))
        dict = mapunion(self.__dict__, locals())
        weave_inline(support_code, norm_code, dict)
        return norm

    def iter_atorbs(self):
        """Iterator over atomic orbitals.

        Yields:
            tuple: (atorb, centre, nshell, N, pl)
                - atorb: AO index
                - centre: centre index
                - nshell: shell index
                - N: radial order
                - pl: polynomial index within shell
        """
        nshell = 0
        atorb = 0
        for centre in range(self.num_centres):
            for _shell in range(self.num_shells_on_centre[centre]):
                for pl in range(num_orbs_per_shelltype[self.shelltype[nshell]]):
                    yield (atorb, centre, nshell, self.order_r_in_shell[nshell], pl)
                    atorb += 1
                nshell += 1

    def cusp_constraint_matrix(self):
        """Construct the cusp condition constraint matrix.

        Returns:
            numpy.ndarray: Matrix of shape (num_centres, num_atorbs) imposing nuclear cusp conditions.
        """
        norm = self.get_norm()
        res = np.asmatrix(np.zeros((self.num_centres, self.num_atorbs)))
        for core in range(self.num_centres):
            atorb_vals = self.eval_atorbs(self.centrepos[core][:, None])[0, :]
            for atorb, centre, nshell, N, _pl in self.iter_atorbs():
                if centre == core:
                    if self.shelltype[nshell] == 1:
                        if N == 0:
                            res[core, atorb] = norm[atorb] * (self.atomcharge[core] - self.zeta[nshell])
                        elif N == 1:
                            res[core, atorb] = norm[atorb]
                else:  # centre != core
                    res[core, atorb] = norm[atorb] * self.atomcharge[core] * atorb_vals[atorb]
        return res

    def cusp_projection_matrix(self):
        """Construct the projection matrix that enforces the cusp condition.

        Returns:
            numpy.ndarray: Projector matrix (num_atorbs, num_atorbs).
        """
        # print "cusp_constraint: ",cusp_constraint
        _U, _S, Vh = np.linalg.svd(self.cusp_constraint_matrix(), full_matrices=False)
        # print "shapes U,S,Vh",U.shape,S.shape,Vh.shape
        P = Vh.T * Vh
        # print "proj shape",cusp_constraint_projector.shape
        # print "proj trace",P.trace()
        # print "proj squarediff", np.linalg.norm(P - P*P)
        Q = np.eye(P.shape[0]) - P
        return Q

    def cusp_fixed_atorbs(self):
        """Determine the atomic orbitals fixed by the cusp constraint.

        Returns:
            numpy.ndarray: Indices of cusp-fixed AOs (num_centres,).
        """
        res = np.zeros(self.num_centres, int)
        for c in range(self.num_centres):
            cidx = np.zeros(self.num_shells)
            cidx[self.idx_first_shell_on_centre[c] : self.idx_first_shell_on_centre[c + 1]] = 1.0
            shell = np.argmax(self.zeta * cidx * (self.shelltype == 1))
            res[c] = num_orbs_per_shelltype[self.shelltype[:shell]].sum()
        return res

    def cusp_enforcing_matrix(self):
        """Construct the linear transformation matrix that enforces cusp conditions.

        Returns:
            numpy.ndarray: Transformation matrix (num_atorbs, num_atorbs).
        """
        cusp_fixed_atorb = self.cusp_fixed_atorbs()
        constraint = self.cusp_constraint_matrix()
        res = constraint + 0.0
        res[:, cusp_fixed_atorb] = 0.0
        U, S, Vh = np.linalg.svd(constraint[:, cusp_fixed_atorb], full_matrices=False)
        tmpinv = Vh.T * np.asmatrix(np.diag(1 / S)) * U.T
        res = -tmpinv * res
        mat = np.asmatrix(np.eye(self.num_atorbs))
        mat[cusp_fixed_atorb, :] = res
        return mat


if __name__ == '__main__':
    sto = stowfn('stowfn.data')
    sto.read_molorbmods('correlation.data')
    points = np.zeros((3, 4))
    points[:, 0] = (-0.19450689, -0.94412413, -0.67370571)
    points[:, :] = points[:, :1]
    points[0, 1] += 0.00317100
    points[1, 2] += 0.00317100
    points[2, 3] += 0.00317100
    val, grad, lap = sto.eval_molorb_derivs(points)
    print('grad analytic:', grad[:, 0])
    print('grad numeric:', (val[1:] - val[0]) / 0.00317100)
#    print sto.get_norm()
#    sto.writefile("stowfn.data.out")



#!/usr/bin/env python3

import re
import sys
from pathlib import Path
from typing import Any, Dict, List, Union

import numpy as np

# Constants for formatting
INT_FIELD_WIDTH = 12
FLOAT_FIELD_WIDTH = 28
STRING_BLOCK_SIZE = 160


class AdfParser:
    """Parser for ADF TAPE21.asc files."""

    def __init__(self, filename: Union[str, Path]) -> None:
        self.filename = Path(filename)
        self.lines: List[str] = []
        self.data: Dict[str, Dict[str, Any]] = {}

    # -----------------------
    # Utility functions
    # -----------------------
    @staticmethod
    def _float_x(x: str) -> float:
        """Convert ADF float string to Python float."""
        x = re.sub(r'E*-', 'E-', x)
        if x.startswith('E'):
            x = x[1:]
        return float(x)

    @staticmethod
    def _split_n(s: str, n: int) -> List[str]:
        """Split string `s` into fixed-length chunks of size `n`."""
        return [s[i : i + n].strip() for i in range(0, len(s), n)]

    @staticmethod
    def _int_x(s: str) -> int:
        """Convert ADF int string to Python int, handling overflow marker."""
        return -(2**31) if s == '**********' else int(s)

    # -----------------------
    # Parsing primitives
    # -----------------------
    def _parse_integers(self, start: int, count: int) -> (np.ndarray, int):
        values: List[int] = []
        i = start
        while len(values) < count:
            values.extend(self._int_x(s) for s in self._split_n(self.lines[i], INT_FIELD_WIDTH))
            i += 1
        return np.array(values, int), i

    def _parse_floats(self, start: int, count: int) -> (np.ndarray, int):
        values: List[float] = []
        i = start
        while len(values) < count:
            values.extend(self._float_x(s) for s in self._split_n(self.lines[i], FLOAT_FIELD_WIDTH))
            i += 1
        return np.array(values, float), i

    def _parse_strings(self, start: int, count: int) -> (List[str], int):
        raw = ''
        i = start
        while len(raw) < count:
            raw += self.lines[i]
            i += 1
        values = [
            raw[STRING_BLOCK_SIZE * j : STRING_BLOCK_SIZE * (j + 1)].strip() for j in range((len(raw) + STRING_BLOCK_SIZE - 1) // STRING_BLOCK_SIZE)
        ]
        return values, i

    def _parse_bools(self, start: int, count: int) -> (List[bool], int):
        values: List[bool] = []
        i = start
        while len(values) < count:
            values.extend({'T': True, 'F': False}[c] for c in self.lines[i])
            i += 1
        return values, i

    def _parse_value(self, typ: int, start: int, count: int) -> (Any, int):
        if typ == 1:
            return self._parse_integers(start, count)
        elif typ == 2:
            return self._parse_floats(start, count)
        elif typ == 3:
            return self._parse_strings(start, count)
        elif typ == 4:
            return self._parse_bools(start, count)
        else:
            raise ValueError(f'Unexpected type {typ}, expected 1..4')

    # -----------------------
    # High-level parsing
    # -----------------------
    def load(self) -> None:
        """Read file contents into memory."""
        try:
            with open(self.filename, encoding='latin-1') as f:
                self.lines = f.read().splitlines()
        except (FileNotFoundError, IsADirectoryError):
            print(f'File {self.filename} could not be read.')
            sys.exit(1)

    def parse(self) -> Dict[str, Dict[str, Any]]:
        """Parse the loaded file into a nested dictionary."""
        if not self.lines:
            self.load()

        i = 0
        self.data = {}

        try:
            while i < len(self.lines):
                group = self.lines[i].strip()
                i += 1
                key = self.lines[i].strip()
                i += 1

                len1, len2, typ = map(int, self.lines[i].split())
                i += 1

                value, i = self._parse_value(typ, i, len2)

                if len1 == 0:
                    i += 1

                self.data.setdefault(group, {})[key] = value

        except Exception:
            print('Parsing error around:')
            for x in range(max(0, i - 3), min(i + 4, len(self.lines))):
                marker = '>>>>' if x == i else '   |'
                print(f'{marker}{self.lines[x]}')
            raise

        return self.data

    # -----------------------
    # Output
    # -----------------------
    def write_dump(self, outfile: Union[str, Path]) -> None:
        """Write human-readable dump of parsed data."""
        if not self.data:
            self.parse()

        with open(outfile, 'w', encoding='latin-1') as f:
            last_group = ''
            for group, keys in self.data.items():
                if group != last_group:
                    f.write(f'\n{group}\n')
                    last_group = group
                for key, value in keys.items():
                    tab = '\t'
                    newline = '\n'
                    valstr = str(value)
                    if newline in valstr:
                        f.write(f'  {key} = {{{len(value)}}}\n' + tab + f'{valstr.replace(chr(10), chr(10) + tab)}\n')
                    else:
                        f.write(f'  {key} = {valstr}\n')


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print('Usage: adfread.py <file.asc>')
        sys.exit(1)

    ascfname = Path(sys.argv[1])
    if ascfname.suffix != '.asc':
        print('Error: input file must end with .asc')
        sys.exit(1)

    parser = AdfParser(ascfname)
    parser.parse()
    parser.write_dump(ascfname.with_suffix('.txt'))



import argparse


def main():
    """Entry point."""
    parser = argparse.ArgumentParser(
        description='Convert ADF TAPE21.asc to CASINO stowfn.data',
        epilog="""
        Examples:
          %(prog)s                         # use defaults: PLOTCUSPS=False, CUSP_ENFORCE=True
          %(prog)s --plot-cusps            # set PLOTCUSPS=True
          %(prog)s --no-cusp-enforce       # set CUSP_ENFORCE=False
          %(prog)s --plot-cusps --no-cusp-enforce
          %(prog)s --dump                  # generate a text dump of the parsed data
        """,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument('--plot-cusps', action='store_true', help='Enable plotting of cusps (default: False)')

    parser.add_argument('--no-cusp-enforce', dest='cusp_enforce', action='store_false', help='Disable cusp enforcement (default: True)')

    parser.add_argument('--dump', action='store_true', help='Generate a text dump file (.txt) of the parsed ADF data (default: False)')

    args = parser.parse_args()
    return args.plot_cusps, args.cusp_enforce, args.dump



#!/usr/bin/env python3.9

# (C) 2008 Norbert Nemec
# This file is part of the CASINO distribution.
# Permission is given to use the script along with the CASINO program and modify
# it for personal use.

import sys

import numpy as np

from adf2stowf import adfread, cli_main, stowfn

############

PLOTCUSPS, CUSP_ENFORCE, DO_DUMP = cli_main.main()

############

parser = adfread.AdfParser('TAPE21.asc')
data = parser.parse()
if DO_DUMP:
    parser.write_dump('TAPE21.txt')

############

General = data['General']
Geometry = data['Geometry']
Properties = data['Properties']
Basis = data['Basis']
Core = data['Core']
Symmetry = data['Symmetry']

############

(Nspins,) = General['nspin']
spin_restricted = Nspins == 1

Nvalence_electrons = int(General['electrons'][0])

(Natoms,) = Geometry['nnuc']
(Natomtypes,) = Geometry['ntyp']
(Ndummies,) = Geometry['nr of dummy atoms']
(Ndummytypes,) = Geometry['nr of dummy atomtypes']

assert Geometry['nr of atoms'] == Natoms + Ndummies
assert Geometry['nr of atomtypes'] == Natomtypes + Ndummytypes

atyp_idx = Geometry['fragment and atomtype index'].reshape(2, Natoms + Ndummies)[1, :] - 1
assert len(atyp_idx) == Natoms + Ndummies
assert np.all(0 <= atyp_idx[0:Natoms])
assert np.all(atyp_idx[0:Natoms] < Natomtypes)
assert np.all(Natomtypes <= atyp_idx[Natoms : Natoms + Ndummies])
assert np.all(atyp_idx[Natoms : Natoms + Ndummies] < Natomtypes + Ndummytypes)
atyp_idx = atyp_idx[:Natoms]

total_charge_per_atomtype = Geometry['atomtype total charge']
atomicnumber_per_atomtype = np.array([int(c) for c in total_charge_per_atomtype])
assert np.all(atomicnumber_per_atomtype[Natomtypes:] == 0)

#####################
#####################

Nharmpoly_per_shelltype = np.array([0, 1, 4, 3, 5, 7])
Ncartpoly_per_shelltype = np.array([0, 1, 0, 3, 6, 10])

harm2cart_per_shelltype = [
    np.eye(1),  # dummy
    np.eye(1),
    np.eye(1),  # dummy
    np.eye(3),
    # from stowfdet code:
    #   poly(5)=xy
    #   poly(6)=yz
    #   poly(7)=zx
    #   poly(8)=(3*zz-r(2)) == 2zz-xx-yy
    #   poly(9)=(xx-yy)
    np.array(
        [
            [0.0, 0.0, 0.0, -1.0, 1.0, +++1],  # x**2
            [1.0, 0.0, 0.0, 0.0, 0.0, +++0],  # x*y
            [0.0, 0.0, 1.0, 0.0, 0.0, +++0],  # x*z
            [0.0, 0.0, 0.0, -1.0, -1.0, +++1],  # y**2
            [0.0, 1.0, 0.0, 0.0, 0.0, +++0],  # y*z
            [0.0, 0.0, 0.0, 2.0, 0.0, +++1],  # z**2
        ]
    ),
    # from stowfdet code:
    #    xx_yy3=xx-3*yy
    #    xx3_yy=3*xx-yy
    #    zz5=5*zz
    #    zz5_rr = zz5-r(2)
    #    poly(10)=(zz5-3*r(2))*z       ! (2*zz-3*(xx+yy))*z
    #    poly(11)=zz5_rr*x             ! (4*zz-(xx+yy))*x
    #    poly(12)=zz5_rr*y             ! (4*zz-(xx+yy))*y
    #    poly(13)=(xx-yy)*z
    #    poly(14)=xy*z
    #    poly(15)=xx_yy3*x
    #    poly(16)=xx3_yy*y
    np.array(
        [
            [0.0, -1.0, 0.0, 0.0, 0.0, 1.0, 0.0, +++1, 0, 0],  # x*x*x
            [0.0, 0.0, -1.0, 0.0, 0.0, 0.0, 3.0, +++0, 1, 0],  # x*x*y
            [-3.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, +++0, 0, 1],  # x*x*z
            [0.0, -1.0, 0.0, 0.0, 0.0, -3.0, 0.0, +++1, 0, 0],  # x*y*y
            [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, +++0, 0, 0],  # x*y*z
            [0.0, 4.0, 0.0, 0.0, 0.0, 0.0, 0.0, +++1, 0, 0],  # x*z*z
            [0.0, 0.0, -1.0, 0.0, 0.0, 0.0, -1.0, +++0, 1, 0],  # y*y*y
            [-3.0, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0, +++0, 0, 1],  # y*y*z
            [0.0, 0.0, 4.0, 0.0, 0.0, 0.0, 0.0, +++0, 1, 0],  # y*z*z
            [2.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, +++0, 0, 1],  # z*z*z
        ]
    ),
]

cart2harm_per_shelltype = [np.linalg.inv(m) for m in harm2cart_per_shelltype]

#####################
# Valence basis set #
#####################

(nbset,) = Basis['nbset']
(nbos,) = Basis['nbos']
nbaspt = Basis['nbaspt'] - 1
assert nbaspt[0] == 0
assert nbaspt[-1] == nbset

Nvalence_shells_per_atomtype = nbaspt[1:] - nbaspt[:-1]
assert np.all(Nvalence_shells_per_atomtype >= 0)

Nvalence_shells_per_centre = Nvalence_shells_per_atomtype[atyp_idx]

############

nqbas = Basis['nqbas']
lqbas = Basis['lqbas']
alfbas = Basis['alfbas']
assert len(nqbas) == nbset
assert len(lqbas) == nbset
assert len(alfbas) == nbset

valence_shelltype = lqbas + 1 + (lqbas > 0)
valence_shelltype_per_atomtype = [valence_shelltype[nbaspt[a] : nbaspt[a + 1]] for a in range(Natomtypes)]

valence_shelltype_per_centre = [valence_shelltype_per_atomtype[at] for at in atyp_idx]

Nvalence_harmbasfns_per_atomtype = [Nharmpoly_per_shelltype[st].sum() for st in valence_shelltype_per_atomtype]
Nvalence_harmbasfns_per_centre = [Nvalence_harmbasfns_per_atomtype[at] for at in atyp_idx]

valence_order_r = nqbas - lqbas - 1
valence_order_r_per_atomtype = [valence_order_r[nbaspt[a] : nbaspt[a + 1]] for a in range(Natomtypes)]

valence_zeta = alfbas
valence_zeta_per_atomtype = [valence_zeta[nbaspt[a] : nbaspt[a + 1]] for a in range(Natomtypes)]

#############

nbptr = Basis['nbptr'] - 1
assert nbptr[0] == 0
assert nbptr[-1] == nbos

Nvalence_cartbasfn_per_atomtype = nbptr[1:] - nbptr[:-1]
assert np.all(Nvalence_cartbasfn_per_atomtype >= 0)

Nvalence_cartbasfn_per_centre = Nvalence_cartbasfn_per_atomtype[atyp_idx]
assert np.sum(Nvalence_cartbasfn_per_centre) == Basis['naos']

#############

bnorm = Basis['bnorm']
assert len(bnorm) == nbos

valence_cartnorm = bnorm
valence_cartnorm_per_atomtype = [valence_cartnorm[nbptr[a] : nbptr[a + 1]] for a in range(Natomtypes)]

#################
# Core basis set
#################

(ncset,) = Core['ncset']
ncorpt = Core['ncorpt'] - 1
assert ncorpt[0] == 0
assert ncorpt[-1] == ncset

Ncore_shells_per_atomtype = ncorpt[1:] - ncorpt[:-1]
assert np.all(Ncore_shells_per_atomtype >= 0)

Ncore_shells_per_centre = Ncore_shells_per_atomtype[atyp_idx]

nrcset = Core['nrcset'].reshape(Natomtypes, 4)
assert np.all(Ncore_shells_per_atomtype == nrcset.sum(axis=1))
assert ncset == nrcset.sum()

############

nqcor = Core['nqcor']
lqcor = Core['lqcor']
alfcor = Core['alfcor']
cornrm = Core['cornrm']

core_shelltype = lqcor + 1 + (lqcor > 0)
core_shelltype_per_atomtype = [core_shelltype[ncorpt[a] : ncorpt[a + 1]] for a in range(Natomtypes)]

core_order_r = nqcor - lqcor - 1
core_order_r_per_atomtype = [core_order_r[ncorpt[a] : ncorpt[a + 1]] for a in range(Natomtypes)]

core_zeta = alfcor
core_zeta_per_atomtype = [core_zeta[ncorpt[a] : ncorpt[a + 1]] for a in range(Natomtypes)]

core_cartnorm = cornrm
core_cartnorm_per_atomtype_per_shell = [core_cartnorm[ncorpt[a] : ncorpt[a + 1]] for a in range(Natomtypes)]
core_cartnorm_per_atomtype = []

for at in range(Natomtypes):
    cn = []
    for s in range(Ncore_shells_per_atomtype[at]):
        if core_shelltype_per_atomtype[at][s] == 1:
            cn += [np.array([core_cartnorm_per_atomtype_per_shell[at][s]])]
        elif core_shelltype_per_atomtype[at][s] == 3:
            cn += [np.array([core_cartnorm_per_atomtype_per_shell[at][s]] * 3)]
        elif core_shelltype_per_atomtype[at][s] == 4:
            raise ValueError('D type fixed core orbitals not yet implemented')
        elif core_shelltype_per_atomtype[at][s] == 5:
            raise ValueError('F type fixed core orbitals not yet implemented')
        else:
            raise ValueError('unknown shell type')
    if len(cn) > 0:
        core_cartnorm_per_atomtype += [np.concatenate(cn)]
    else:
        core_cartnorm_per_atomtype += [np.zeros([0])]

############
############

Nshells_per_centre = Nvalence_shells_per_centre + Ncore_shells_per_centre

shelltype_per_centre = [np.concatenate([core_shelltype_per_atomtype[at], valence_shelltype_per_atomtype[at]]) for at in atyp_idx]
order_r_per_centre = [np.concatenate([core_order_r_per_atomtype[at], valence_order_r_per_atomtype[at]]) for at in atyp_idx]
zeta_per_centre = [np.concatenate([core_zeta_per_atomtype[at], valence_zeta_per_atomtype[at]]) for at in atyp_idx]

for c in range(Natoms):
    assert len(shelltype_per_centre[c]) == Nshells_per_centre[c]
    assert len(order_r_per_centre[c]) == Nshells_per_centre[c]
    assert len(zeta_per_centre[c]) == Nshells_per_centre[c]

###############
# coefficients
###############

Nharmbasfns_per_centre = [Nharmpoly_per_shelltype[st].sum() for st in shelltype_per_centre]
Nharmbasfns = np.sum(Nharmbasfns_per_centre)

assert np.all(Nvalence_cartbasfn_per_centre == np.array([Ncartpoly_per_shelltype[st].sum() for st in valence_shelltype_per_centre]))
Nvalence_cartbasfn = np.sum(Nvalence_cartbasfn_per_centre)

####################
# valence orbitals #
####################

(nsym,) = Symmetry['nsym']
symlab = Symmetry['symlab']
assert len(symlab) == nsym
norb = Symmetry['norb']
assert len(norb) == nsym


def select_coeff(sp):
    """Select valence molecular orbital coefficients for a given spin channel.

    Args:
       sp (int): Spin index (0 for alpha, 1 for beta)

    Returns:
       valence_molorb_cart_coeff (np.ndarray): Array of shape (n_orbs, n_basis),
           containing the selected valence orbital coefficients in Cartesian basis.
    """
    X = ['A', 'B'][sp]  # Label for spin channel: "A" or "B"

    # Lists to store valence orbital data
    valence_molorb_cart_coeff = []
    valence_molorb_occupation = []
    valence_molorb_eigenvalue = []
    # Dictionary to store leftover partial occupations
    partial_occupations = {}
    # Loop over all symmetries
    for sym in range(nsym):
        Section = data[symlab[sym]]
        # Number of orbitals for this spin and symmetry
        (nmo_X,) = Section['nmo_' + X]
        assert nmo_X == norb[sym]
        # Fractional occupations for each orbital
        froc_X = Section['froc_' + X]
        assert len(froc_X) == norb[sym]
        # Skip if all occupations are zero
        if np.all(froc_X == 0.0):
            continue
        # Indices of basis functions for this symmetry
        npart = Section['npart'] - 1
        # Extract molecular orbital coefficients and eigenvalues
        Eigen_Bas_X = Section['Eigen-Bas_' + X].reshape([nmo_X, len(npart)])
        eps_X = Section['eps_' + X].reshape([nmo_X])
        # Loop over all orbitals
        for o in range(nmo_X):
            eigv = eps_X[o]

            valence_molorb_eigenvalue += [eigv]
            occ = froc_X[o]
            # Add any leftover partial occupation for this eigenvalue
            if eigv in partial_occupations:
                occ += partial_occupations.pop(eigv)
            # Check if orbital is considered "occupied"
            if occ + 1e-8 >= 2.0 / Nspins:
                valence_molorb_occupation += [1]
                occ -= 2.0 / Nspins
                # Construct coefficient vector in Cartesian basis
                coeff = np.zeros((Nvalence_cartbasfn,))
                coeff[npart] = Eigen_Bas_X[o, :]
                valence_molorb_cart_coeff += [coeff]
            else:
                valence_molorb_occupation += [0]
            # Store leftover fractional occupation
            if occ > 1e-8:
                partial_occupations[eigv] = occ
    # Print any leftover partial occupations
    for k, v in iter(partial_occupations.items()):
        print('spin=', sp, ': leftover partial occupation at E=', k, ': ', v)
    # Sanity check: should be no leftover occupation
    assert np.sum(len(p) for p in partial_occupations) == 0
    # Nmolorbs_total = len(valence_molorb_eigenvalue)
    # Number of occupied valence orbitals
    Nmolorbs_occup = len(valence_molorb_cart_coeff)

    assert np.sum(valence_molorb_occupation) == Nmolorbs_occup
    # Convert lists to NumPy arrays
    valence_molorb_occupation = np.array(valence_molorb_occupation)
    valence_molorb_eigenvalue = np.array(valence_molorb_eigenvalue)
    valence_molorb_cart_coeff = np.array(valence_molorb_cart_coeff)

    # Ensure 2D shape even if only one orbital exists (e.g., hydrogen)
    if valence_molorb_cart_coeff.ndim == 1:
        valence_molorb_cart_coeff = valence_molorb_cart_coeff.reshape(1, -1)

    # Identify occupied and unoccupied orbitals
    occupied = valence_molorb_occupation[:] == 1
    occidx = valence_molorb_eigenvalue[occupied]
    unoccidx = valence_molorb_eigenvalue[~occupied]
    # Check HOMO-LUMO ordering (warning if HOMO > LUMO)
    if len(occidx) > 0 and len(unoccidx) > 0:
        HOMO = max(occidx)
        LUMO = min(unoccidx)
        if HOMO > LUMO:
            print('Warning: HOMO > LUMO (may happen in some cases)')
    # Keep only occupied eigenvalues
    valence_molorb_eigenvalue = valence_molorb_eigenvalue[occupied]
    # Sanity check: number of orbitals matches number of coefficients
    assert len(valence_molorb_eigenvalue) == Nmolorbs_occup
    # Sort orbitals by eigenvalue
    order = valence_molorb_eigenvalue.argsort()
    valence_molorb_cart_coeff = valence_molorb_cart_coeff[order, :]

    return valence_molorb_cart_coeff


valence_molorb_cart_coeff = [select_coeff(sp) for sp in range(Nspins)]
Nvalence_molorbs = np.array([c.shape[0] for c in valence_molorb_cart_coeff])

assert np.sum(Nvalence_molorbs) * (3 - Nspins) == Nvalence_electrons

##############################

cart2harm_matrix = np.zeros((Nharmbasfns, Nvalence_cartbasfn))
cart2harm_constraint = []
i, j = 0, 0
for c in range(Natoms):
    at = atyp_idx[c]
    for st in core_shelltype_per_atomtype[at]:
        i += Nharmpoly_per_shelltype[st]

    for st in valence_shelltype_per_atomtype[at]:
        if st == 1:  # S shell
            cart2harm_matrix[i, j] = 1.0
            i += 1
            j += 1
        elif st == 3:  # P shell
            cart2harm_matrix[i : i + 3, j : j + 3] = np.eye(3)
            i += 3
            j += 3
        elif st == 4:  # D shell
            cart2harm_matrix[i : i + 5, j : j + 6] = cart2harm_per_shelltype[st][:5, :]
            constraint = np.zeros([1, Nvalence_cartbasfn])
            constraint[:, j : j + 6] = cart2harm_per_shelltype[st][5:, :]
            cart2harm_constraint += [constraint]
            i += 5
            j += 6

        elif st == 5:  # F shell
            cart2harm_matrix[i : i + 7, j : j + 10] = cart2harm_per_shelltype[st][:7, :]
            constraint = np.zeros([3, Nvalence_cartbasfn])
            constraint[:, j : j + 10] = cart2harm_per_shelltype[st][7:, :]
            cart2harm_constraint += [constraint]
            i += 7
            j += 10

assert i == Nharmbasfns
assert j == Nvalence_cartbasfn

if len(cart2harm_constraint) > 0:
    cart2harm_constraint = np.concatenate(cart2harm_constraint, axis=0)

valence_molorb_harm_coeff = [np.zeros((Nharmbasfns, Nvalence_molorbs[sp])) for sp in range(Nspins)]

for sp in range(Nspins):
    for m in range(Nvalence_molorbs[sp]):
        valence_molorb_harm_coeff[sp][:, m] = cart2harm_matrix @ valence_molorb_cart_coeff[sp][m, :]
        if len(cart2harm_constraint) > 0:
            violation = cart2harm_constraint @ valence_molorb_cart_coeff[sp][m, :]
            absviolation = np.sqrt(np.sum(np.abs(violation**2)))
            if absviolation > 1e-5:
                print('WARNING: cartesian to harmonic conversion: spin #%i, orb #%i ' 'violated by %g' % (sp, m, absviolation))


#######################
# fixed core orbitals #
#######################

nrcorb = Core['nrcorb'].reshape(Natomtypes, 4)
ccor = Core['ccor']
Nccor_per_atomtype = (nrcset * nrcorb).sum(axis=1)
assert len(ccor) == Nccor_per_atomtype.sum()
ccor_per_atomtype = np.array_split(ccor, np.cumsum(Nccor_per_atomtype))[:-1]

Ncoremolorbs_per_atomtype = (nrcorb * np.array([1, 3, 5, 7])[None, :]).sum(axis=1)
Ncoremolorbs_per_centre = Ncoremolorbs_per_atomtype[atyp_idx]


Ncore_molorbs = Ncoremolorbs_per_centre.sum()

core_molorb_coeff = np.zeros((Nharmbasfns, Ncore_molorbs))

molorb = 0

for a in range(Natoms):
    at = atyp_idx[a]
    first_harmbasfn = np.sum(Nharmbasfns_per_centre[:a])
    Ncore_harmbasfns = np.sum(Nharmpoly_per_shelltype[st].sum() for st in core_shelltype_per_atomtype[at])
    core_coeff = np.zeros([Ncore_harmbasfns])
    ccor_per_shell = np.array_split(ccor_per_atomtype[at], np.cumsum((nrcset * nrcorb)[at, :]))[:-1]

    for shell in range(nrcorb[at, 0]):  # S core shells
        core_coeff[:] = 0.0
        core_coeff[0 : nrcset[at, 0]] = ccor_per_shell[0][nrcset[at, 0] * shell : nrcset[at, 0] * (shell + 1)]
        core_molorb_coeff[first_harmbasfn : first_harmbasfn + Ncore_harmbasfns, molorb] = core_coeff
        molorb += 1

    for shell in range(nrcorb[at, 1]):  # P core shells
        for i in range(3):
            core_coeff[:] = 0.0
            offset = nrcset[at, 0]
            core_coeff[offset + i : offset + nrcset[at, 1] * 3 : 3] = ccor_per_shell[1][nrcset[at, 1] * shell : nrcset[at, 1] * (shell + 1)]
            core_molorb_coeff[first_harmbasfn : first_harmbasfn + Ncore_harmbasfns, molorb] = core_coeff
            molorb += 1
    for shell in range(nrcorb[at, 2]):  # D core shells
        for i in range(5):
            core_coeff[:] = 0.0
            offset = nrcset[at, 0] + nrcset[at, 1]
            core_coeff[offset + i : offset + nrcset[at, 2] * 5 : 5] = ccor_per_shell[2][nrcset[at, 2] * shell : nrcset[at, 2] * (shell + 1)]
            core_molorb_coeff[first_harmbasfn : first_harmbasfn + Ncore_harmbasfns, molorb] = core_coeff
            molorb += 1
    for shell in range(nrcorb[at, 3]):  # F core shells
        for i in range(7):
            core_coeff[:] = 0.0
            offset = nrcset[at, 0] + nrcset[at, 1] + nrcset[at, 2]
            core_coeff[offset + i : offset + nrcset[at, 2] * 7 : 7] = ccor_per_shell[3][nrcset[at, 3] * shell : nrcset[at, 3] * (shell + 1)]
            core_molorb_coeff[first_harmbasfn : first_harmbasfn + Ncore_harmbasfns, molorb] = core_coeff
            molorb += 1

assert molorb == Ncore_molorbs

############
############

Nmolorbs = np.array([Ncore_molorbs + Nvalence_molorbs[sp] for sp in range(Nspins)])
coeff = [np.concatenate([core_molorb_coeff, valence_molorb_harm_coeff[sp]], axis=1) for sp in range(Nspins)]

if False:
    print('molorb sparsity:')
    for sp in range(Nspins):
        for i in range(Nmolorbs[sp]):
            print(''.join(np.array(['.', 'X'])[(coeff[sp][:, i] != 0.0) * 1]))


############
############

norm_per_centre = [np.concatenate([core_cartnorm_per_atomtype[at], valence_cartnorm_per_atomtype[at]]) for at in atyp_idx]

norm_per_harmbasfn = np.concatenate(norm_per_centre)
# print(norm_per_centre)
# print(Nharmbasfns)
# assert len(norm_per_harmbasfn) == Nharmbasfns

############
############


sto = stowfn.stowfn()

sto.num_atom = Natoms

(sto.title,) = General['title']
sto.code = 'ADF'
sto.periodicity = 0
sto.spin_unrestricted = not spin_restricted
sto.nuclear_repulsion_energy = 0.0
sto.atomcharge = total_charge_per_atomtype[atyp_idx]
assert len(sto.atomcharge) == Natoms

eionion = 0.0
if Natoms > 1:
    adist = Geometry['Atomic Distances'].reshape(Natoms + 1, Natoms + 1)[1:, 1:]
    for i in range(Natoms):
        assert adist[i, i] == 0.0
        for j in range(i):
            assert adist[i, j] == adist[j, i]
            assert adist[i, j] > 0.0
            eionion += sto.atomcharge[i] * sto.atomcharge[j] / adist[i, j]
sto.nuclear_repulsion_energy = eionion / Natoms

sto.num_elec = Nvalence_electrons + 2 * Ncore_molorbs
sto.atompos = Geometry['xyz'].reshape(Natoms + Ndummies, 3)[:Natoms, :]
sto.atomnum = atomicnumber_per_atomtype[atyp_idx]

###############
# basis set

sto.num_centres = int(Natoms)
sto.centrepos = Geometry['xyz'].reshape(Natoms + Ndummies, 3)[:Natoms, :]
sto.num_shells = np.sum(Nshells_per_centre)

sto.idx_first_shell_on_centre = np.array([0] + list(np.cumsum(Nshells_per_centre)))

sto.shelltype = np.concatenate(shelltype_per_centre)
sto.order_r_in_shell = np.concatenate(order_r_per_centre)
sto.zeta = np.concatenate(zeta_per_centre)

sto.num_atorbs = Nharmbasfns
sto.num_molorbs = Nmolorbs
sto.coeff = [c.T for c in coeff]
sto.footer = ''

sto.check_and_normalize()

# check norm

if False:
    print('Norm-ADF (cartesian)')
    print(norm_per_harmbasfn)
    print('Norm-computed (minimal basis)')
    print(sto.get_norm())
    # print(norm_per_harmbasfn - sto.get_norm())

np.set_printoptions(suppress=True)

# assert np.all(np.abs(norm_per_harmbasfn - sto.get_norm()) < 1e-13)

cusp_fixed_atorbs = sto.cusp_fixed_atorbs()
cusp_constraint = sto.cusp_constraint_matrix()
# print("cusp_constraint_matrix:")
# print(cusp_constraint)
# cusp_projection = sto.cusp_projection_matrix()
cusp_enforcing = sto.cusp_enforcing_matrix()

print('Molorb values at nuclei before applying cusp constraint:')
print(sto.eval_molorbs(sto.atompos.transpose()))
# Initialize a list of boolean masks, one per spin channel.
# Each mask has length equal to the number of molecular orbitals (Nmolorbs[sp]) for that spin.
# Initially, all values are False (no violations detected yet).
fixed = [np.zeros(Nmolorbs[sp], bool) for sp in range(Nspins)]

for sp in range(Nspins):
    for i in range(Nmolorbs[sp]):
        # Compute the cusp constraint violation for orbital i of spin sp
        constraint_violation = cusp_constraint @ coeff[sp][:, i]
        # If any component of the violation is larger than the tolerance (1e-9),
        # we mark this orbital as "fixed" (problematic) and handle it.
        if np.any(np.abs(constraint_violation) > 1e-9):
            fixed[sp][i] = True
            print('spin #%i, orb #%i - constraint violation by:' % (sp, i), constraint_violation)
            if CUSP_ENFORCE:
                # Show original coefficients for the constrained atomic orbitals
                print('    original coefficients:    ', coeff[sp][cusp_fixed_atorbs, i])
                # Projected coefficients (alternative approach, commented out)
                # projected_coeff = cusp_projection.A @ coeff[:,i]
                # print("    proj coeff:",projected_coeff)
                # print("    after projection       :", cusp_constraint @ projected_coeff)
                # Apply the cusp enforcing projection to fix the coefficients
                enforced_coeff = cusp_enforcing.A @ coeff[sp][:, i]
                print('    constrained coefficients: ', enforced_coeff[cusp_fixed_atorbs])
                # Replace the original coefficients with the enforced (corrected) ones
                coeff[sp][:, i] = enforced_coeff
                # Re-check that constraint violation is now within the stricter tolerance
                constraint_violation = cusp_constraint @ coeff[sp][:, i]
                assert np.all(np.abs(constraint_violation) < 1e-8)
                # print("    after enforcing        :", cusp_constraint @ enforced_coeff)

if PLOTCUSPS:
    # Build a z-axis line through each atom from -0.5 to 0.5 (relative units)
    z = np.linspace(-0.5, 0.5, 501)
    r = [np.zeros((3, len(z))) + sto.atompos[at, :][:, None] for at in range(sto.num_atom)]
    for ir in r:
        ir[2, :] += z
    # Print the boolean masks (which orbitals are marked as violating cusp conditions)
    print(fixed)
    val_pre = [[sto.eval_molorbs(ir, spin=sp)[:, fixed[sp]] for sp in range(Nspins)] for ir in r]
    lap_pre = [[sto.eval_molorb_derivs(ir, spin=sp)[2][:, fixed[sp]] for sp in range(Nspins)] for ir in r]

sto.coeff = [c.T for c in coeff]
sto.check_and_normalize()

if PLOTCUSPS:
    val_post = [[sto.eval_molorbs(ir, spin=sp)[:, fixed[sp]] for sp in range(Nspins)] for ir in r]
    lap_post = [[sto.eval_molorb_derivs(ir, spin=sp)[2][:, fixed[sp]] for sp in range(Nspins)] for ir in r]

if CUSP_ENFORCE:
    print('Molorb values at nuclei after applying cusp constraint:')
    print(sto.eval_molorbs(sto.atompos.transpose()))
    # assert np.all(np.abs(norm_per_harmbasfn - sto.get_norm()) < 1e-13)

sto.writefile('stowfn.data')

if PLOTCUSPS:
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        print('The PLOTCUSPS feature requires the matplotlib library, which could not be found.')
        sys.exit()
    Natom = sto.num_atom
    # Create a 2 x Natom grid of subplots
    # Top row: wavefunction values (val)
    # Bottom row: local energies (eloc)
    fig, axes = plt.subplots(2, Natom, figsize=(4 * Natom, 4))
    # If only one atom, reshape axes into 2D array for consistency
    if Natom == 1:
        axes = np.array([axes]).reshape(2, 1)
    axval = [axes[0, at] for at in range(Natom)]
    axeloc = [axes[1, at] for at in range(Natom)]
    for at in range(Natom):
        eloc_min = 1e8
        eloc_max = -1e8
        for sp in range(Nspins):
            for i in range(np.sum(fixed[sp])):
                vpre = val_pre[at][sp][:, i]  # wavefunction before correction
                vpost = val_post[at][sp][:, i]  # wavefunction after correction
                sgn = np.sign(vpre[len(vpre) // 2])  # sign normalization
                # Plot wavefunction before and after correction
                (pl,) = axval[at].plot(z, sgn * vpre, '--')
                axval[at].plot(z, sgn * vpost, color=pl.get_color())
                # Plot Laplacian of the wavefunction + Coulomb term (without normalization).
                # pl, = axeloc[at].plot(z, lap_pre[at][sp][:,i] + sto.atomnum[at]/z, '--')
                # Compute local energy before and after correction:
                # E_loc = -0.5 * (Laplacian / wavefunction) - Z / |r|
                eloc_pre = -0.5 * lap_pre[at][sp][:, i] / val_pre[at][sp][:, i] - sto.atomnum[at] / np.abs(z)
                eloc_post = -0.5 * lap_post[at][sp][:, i] / val_post[at][sp][:, i] - sto.atomnum[at] / np.abs(z)
                # Plot local energy before and after correction
                axeloc[at].plot(z, eloc_pre, '--', color=pl.get_color())
                axeloc[at].plot(z, eloc_post, '-', color=pl.get_color())
                # Track min/max values for axis scaling
                eloc_min = min(
                    eloc_min,
                    eloc_post[0],
                    eloc_post[-1],
                    eloc_post[len(eloc_post) // 2 - 1],
                    eloc_post[len(eloc_post) // 2 + 1],
                )
                eloc_max = max(
                    eloc_min,
                    eloc_post[0],
                    eloc_post[-1],
                    eloc_post[len(eloc_post) // 2 - 1],
                    eloc_post[len(eloc_post) // 2 + 1],
                )
        # Expand y-limits around the middle value for better visualization
        eloc_mid = (eloc_min + eloc_max) / 2
        eloc_min = (eloc_min - eloc_mid) * 1.5 + eloc_mid
        eloc_max = (eloc_max - eloc_mid) * 1.5 + eloc_mid
        # Set axis ranges
        axval[at].set_xlim(z[0], z[-1])
        axeloc[at].set_xlim(z[0], z[-1])
        axeloc[at].set_ylim(eloc_min, eloc_max)
    # Adjust layout and save figure
    fig.tight_layout()
    fig.savefig('cusp_constraint.svg')


def main():
    """Entry point."""
